from atlantiscore.types.evm.address import (  # noqa
    ADDRESS_BYTE_COUNT,
    EXAMPLE_ADDRESS_STRING,
    EVMAddress,
)
from atlantiscore.types.evm.base import ByteEncoding, LiteralByteEncoding  # noqa
from atlantiscore.types.evm.transaction import (  # noqa
    EXAMPLE_TRANSACTION_HASH_STRING,
    TRANSACTION_HASH_BYTE_COUNT,
    EVMTransactionHash,
)



class InvalidByteEncoding(RuntimeError):
    def __init__(self, invalid_value: any) -> None:
        super().__init__(
            f"{invalid_value} is not transformable to a valid byte sequence."
        )


class InvalidEVMAddress(InvalidByteEncoding):
    def __init__(self, invalid_value: any) -> None:
        super().__init__(invalid_value)
        self.args = (f"{invalid_value} is not transformable to a valid EVM address.",)


class InvalidEVMTransactionHash(InvalidByteEncoding):
    def __init__(self, invalid_value: any) -> None:
        super().__init__(invalid_value)
        self.args = (
            f"{invalid_value} is not transformable to a valid EVM transaction hash.",
        )



from dataclasses import dataclass, fields
from typing import Any, Iterator

from eth_account.messages import encode_typed_data
from eth_utils import keccak

from atlantiscore.lib.string import snake_to_camel_case
from atlantiscore.types import EVMAddress, PrivateKey


@dataclass
class EIP712Domain:
    name: str | None = None
    version: str | None = None
    chain_id: int | None = None
    verifying_contract: EVMAddress | None = None

    def types(self) -> list[dict]:
        types = []
        if self.name is not None:
            types.append(create_type_info("name", "string"))
        if self.version is not None:
            types.append(create_type_info("version", "string"))
        if self.chain_id is not None:
            types.append(create_type_info("chainId", "uint256"))
        if self.verifying_contract is not None:
            types.append(create_type_info("verifyingContract", "address"))
        return types

    def __iter__(self) -> Iterator[tuple[str, Any]]:
        for field in fields(EIP712Domain):
            attr = field.name
            value = getattr(self, attr)
            if value is None:
                continue

            yield (snake_to_camel_case(attr), value)


class EIP712Signer:
    private_key: PrivateKey
    domain: EIP712Domain

    def __init__(self, private_key: PrivateKey, domain: EIP712Domain) -> None:
        self.private_key = private_key
        self.domain = domain

    def sign(
        self,
        message: dict,
        primary_type: str,
        additional_types: dict = {},
    ) -> bytes:
        domain_data = dict(self.domain)
        if verifying_contract := domain_data.get("verifyingContract"):
            domain_data["verifyingContract"] = str(verifying_contract)
        payload = {
            "types": {"EIP712Domain": self.domain.types(), **additional_types},
            "primaryType": primary_type,
            "domain": domain_data,
            "message": message,
        }
        encoded_data = encode_typed_data(full_message=payload)
        return self.private_key.sign_recoverable(
            b"\x19" + encoded_data.version + encoded_data.header + encoded_data.body,
            hasher=keccak,
        )


def create_type_info(name: str, type: str) -> dict[str, str]:
    return {"name": name, "type": type}



from typing import Self

from coincurve import GLOBAL_CONTEXT, Context, PrivateKey as CoinCurvePrivateKey
from coincurve.utils import hex_to_bytes
from eth_utils import keccak

from atlantiscore.types.evm import EVMAddress

NUMBER_OF_BYTES_IN_ADDRESS = 20


class PrivateKey(CoinCurvePrivateKey):
    @property
    def public_address(self) -> EVMAddress:
        public_key = self.public_key.format(compressed=False)[1:]
        return EVMAddress(keccak(public_key)[-NUMBER_OF_BYTES_IN_ADDRESS:])

    @classmethod
    def from_der(cls, *args, **kwargs) -> Self:
        return cls.from_hex(CoinCurvePrivateKey.from_der(*args, **kwargs).to_hex())

    @classmethod
    def from_hex(cls, hexed: str, context: Context = GLOBAL_CONTEXT) -> Self:
        return PrivateKey(hex_to_bytes(hexed), context)

    @classmethod
    def from_int(cls, *args, **kwargs) -> Self:
        return cls.from_hex(CoinCurvePrivateKey.from_int(*args, **kwargs).to_hex())

    @classmethod
    def from_pem(cls, *args, **kwargs) -> Self:
        return cls.from_hex(CoinCurvePrivateKey.from_pem(*args, **kwargs).to_hex())



import asyncio

from web3.eth import Eth
from web3.exceptions import TimeExhausted, TransactionNotFound


async def wait_for_transaction_receipt(
    eth: Eth,
    transaction_hash: str,
    timeout_in_seconds: int = 120,
    poll_latency: float = 0.1,
) -> dict:
    try:
        return await asyncio.wait_for(
            _wait_for_transaction_receipt_indefinetly(
                eth, transaction_hash, poll_latency
            ),
            timeout_in_seconds,
        )
    except asyncio.TimeoutError:
        raise TimeExhausted("Transaction receipt not found within the given time limit")


async def _wait_for_transaction_receipt_indefinetly(
    eth: Eth,
    transaction_hash: str,
    poll_latency: float = 0.1,
) -> dict:
    while True:
        try:
            return eth.get_transaction_receipt(transaction_hash)
        except TransactionNotFound:
            await asyncio.sleep(poll_latency)
